#include "Cylinder.h"
#include <algorithm>
#include <iostream>

#ifdef OPENIV
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoLabel.h>
#endif

// by Wen Jiang, June 2000

/***************************************************************************************
		Cylinder class
****************************************************************************************/
Cylinder::Cylinder()
{
	cen.setXYZ(0,0,0);
	axis.setXYZ(0,0,1);
	//xc = yc = zc = alt = az = 0.0;	
	radius = 2.5; 
	length = 5.4;	
	weight = 0.9;	
	startid=-1;
	endid=-1;
}
 
Cylinder::Cylinder(const Point& center, const Vect& axisdir, float len, float r, float w)
{
	cen=center;
	axis=axisdir;
	axis.normalize();
	radius = r;	
	length = len;	
	weight = w;	
	startid=-1;
	endid=-1;
}

float	Cylinder::Xc() 
{
	return cen.X();
}

float	Cylinder::Yc() 
{
	return cen.Y();
}

float	Cylinder::Zc()
{
	return cen.Z();
}

float	Cylinder::Alt()
{
	return axis.Alt();
}

float	Cylinder::Az() 
{
	return axis.Az();
}

Point	Cylinder::Center()
{
	return cen;
}

Vect	Cylinder::Axis()
{
	return axis;
}

Quater	Cylinder::QuaterForAxis()
{
	return axis.getQuater();
}

float	Cylinder::Radius() 
{
	return radius;
}

float	Cylinder::Length() 
{
	return length;
}

float	Cylinder::Weight() 
{
	return weight;
}

int		Cylinder::StartID()
{
	return startid;
}

int		Cylinder::EndID()
{
	return endid;
}

string	Cylinder::Label()
{	
	return label;
}

void	Cylinder::mirror()
{
	axis.setX(-axis.X());
	cen.setX(-cen.X());
}

void	Cylinder::Transform(Quater& rot, Vect& trans, Vect& center)
{	
	Euler euler=rot.toEuler();
	Transform(euler,trans,center);
}

void	Cylinder::Transform(Euler& rot, Vect& trans, Vect& center)
{	// transform: rotate then translate
	Euler axiseuler=axis.euler();
	Euler axiseuler2=rot*axiseuler;		
	axis=Vect(0,0,1).rotate(axiseuler2);	
	cen=(cen-center).rotate(rot)+center+trans;	
	
	/*Transform3D prerot=Transform3D(rot);	//prerot.print();
	Transform3D pretrans=Transform3D(trans);	//pretrans.print();
	Transform3D selfrot=Transform3D(axiseuler);	//selfrot.print();
	Transform3D selftrans=Transform3D(cen);	//selftrans.print();
	
	Transform3D tempt=selfrot*prerot;	//tempt.print();
	Euler axiseuler2=tempt.getRotation();	//axiseuler2.print("Transform3D");
	Euler axiseuler3=axiseuler*rot;		//axiseuler3.print("Euler      ");
	axis=Vect(0,0,1).rotate(axiseuler2);	
	cen=(cen-center).rotate(rot)+center+trans;	
	*/
}

void	Cylinder::setCenter(float x, float y, float z) 
{
	cen.setXYZ(x,y,z);		
}

void	Cylinder::setAxis(const Vect& axisdir)
{
	axis=axisdir;
	axis.normalize();
}

void	Cylinder::setAxis(Euler& axisdir)
{
	axis=Vect(0,0,1).rotate(axisdir);
	axis.normalize();
}

void	Cylinder::setAxis(Euler* axisdir)
{
	axis=Vect(0,0,1).rotate(axisdir);
	axis.normalize();
}

void	Cylinder::setRadius(float r) 
{
	radius = r;
}

void	Cylinder::setLength(float l) 
{
	length = l;
}

void	Cylinder::setWeight(float w) 
{
	if(w>1 || w<0) {
		fprintf(stderr,"ERROR: w=%f, but it must be in [0,1] for \"Cylinder::setWeight(float w)\"\n",w);
		exit(1);
	}
	weight = w;
}

void	Cylinder::setStartID(int i)
{
	startid=i;
}

void	Cylinder::setEndID(int i)
{
	endid=i;
}

void	Cylinder::setLabel(const string& l)
{
	label=l;
}

void	Cylinder::print(int id, FILE *out)
{
	if (out == stdout) {
		if (id < 0)
			fprintf(out, "Helix:    \t");
		else 
			fprintf(out, "Helix %3d: ", id);
		fprintf(out, "Center=( %5.1f , %5.1f , %5.1f )  Axis=( %6.3f , %6.3f , %6.3f )  Radius= %3.1f  Length= %3.1f  Weight= %.2f\n",
		     cen.X(), cen.Y(), cen.Z(), axis.X(), axis.Y(), axis.Z(), radius, length, weight);
	} else {
		Euler tmp=Euler(PI/2., 0, 0)*axis.euler();		
		Quater quater(tmp);		
		fprintf(out, "%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n", cen.X(),cen.Y(),cen.Z(), 
				quater.nx, quater.ny, quater.nz, quater.Q, radius, length);
	}
}

void	Cylinder::toInventor(FILE *out, int id, int indent, bool with_head)
{
	char	tab[80];
	int	i;
	for (i = 0; i < indent; i++)
		tab[i] = '\t';
	tab[i] = 0;		// setup the string to include indent of table char for formatting
	if (with_head)
		fprintf(out, "#Inventor V2.1 ascii\n\nSeparator {\n");
	
	if(label.empty())
		fprintf(out, "%sSeparator {\n\t\tLabel {\n\t\t\tlabel\t\"helix %d coeff= %f\"\n\t\t}\n", tab, id, weight);
	else
		fprintf(out, "%sSeparator {\n\t\tLabel {\n\t\t\tlabel\t\"helix %s\"\n\t\t}\n", tab, label.c_str());
	
	fprintf(out, "%s\tBaseColor {\n\t\t\trgb %f %f %f\n\t\t}\n", tab, weight*0, weight, (1-weight));

	fprintf(out, "%s\tTransform {\n", tab);

	//Euler tmp=Euler(PI/2., 0, 0)*axis.euler();		
	Euler tmp=Euler(PI/2., 0, 0)*axis.euler().inverse();		
	Quater quater(tmp);		
	float	rx = quater.nx,  ry = quater.ny,  rz = quater.nz, 	q = quater.Q;
	fprintf(out, "%s\t\trotation\t%f %f %f %f\n", tab, rx, ry, rz, q);
	fprintf(out, "%s\t\ttranslation\t%f %f %f\n", tab, cen.X(), cen.Y(), cen.Z());
	fprintf(out, "%s\t}\n", tab);

	fprintf(out, "%s\tCylinder {\n\t\t\tradius\t%f\n\t\t\theight\t%f\n\t\t}\n\t}\n", tab, radius, length);

	if (with_head)
		fprintf(out, "}\n");
}


void	Cylinder::toDejaVu(FILE *out, int id)
{
	float	xe1, xe2, ye1, ye2, ze1, ze2;
	xe1 = cen.X() - length / 2 * axis.X();
	ye1 = cen.Y() - length / 2 * axis.Y();
	ze1 = cen.Z() - length / 2 * axis.Z();
	xe2 = cen.X() + length / 2 * axis.X();
	ye2 = cen.Y() + length / 2 * axis.Y();
	ze2 = cen.Z() + length / 2 * axis.Z();
	int	resnum = round(length / 1.6+1);

	if(startid<=0 || endid<=0) {
		startid=id*100;
		endid=startid+resnum-1;
	}
	
	fprintf(out, "ALPHA 'A%d' '%d' '%d' %d %f %f %f %f %f %f\n", id, startid,
	    endid, resnum, xe1, ye1, ze1, xe2, ye2, ze2);
}

void	Cylinder::toCosec(FILE *out, int id)
{
	float	xe1, xe2, ye1, ye2, ze1, ze2;
	int	resnum = round(length / 1.6+1);
	if(startid<=0 || endid<=0) {
		startid=id*100;
		endid=startid+resnum-1;
	}
	
	xe1 = cen.X() - length / 2 * axis.X();
	ye1 = cen.Y() - length / 2 * axis.Y();
	ze1 = cen.Z() - length / 2 * axis.Z();
	xe2 = cen.X() + length / 2 * axis.X();
	ye2 = cen.Y() + length / 2 * axis.Y();
	ze2 = cen.Z() + length / 2 * axis.Z();
	fprintf(out,"%4d  h %7.3f %7.3f %7.3f   %7.3f %7.3f %7.3f  %7.3f %4d %4d  - %4d \n",
		id,xe1,ye1,ze1,xe2,ye2,ze2,length,resnum,startid,endid);
	
}

float	Cylinder::Distance2Point(float x, float y, float z)
{
	return((cen-Point(x,y,z)).length());
}


float	Cylinder::CenterDistance(Cylinder *next)
{
	return(Distance2Point(next->Xc(), next->Yc(), next->Zc()));
}


float	Cylinder::NormDistance(Cylinder *next)
{
	float x1=cen.X(), y1=cen.Y(), z1=cen.Z(); 
	Point cen2=next->Center();
	Vect axis2=next->Axis();
	float x2=cen2.X(), y2=cen2.Y(), z2=cen2.Z(); 
	float denom=sqrt(hypot3(x1*y2-x2*y1, y1*z2-y2*z1, z1*x2-z2*x1));
	Vect tmpv=axis.cross(axis2);	
	return fabs((cen2-cen).inner(tmpv))/denom;
}


float	Cylinder::Angle2Vect(Vect& v)
{
	float tmp=axis.inner(v)/(axis.length()*v.length());
	if(tmp>1)tmp=1;
	else if(tmp<-1)tmp=-1;
	float da=acos(tmp);
	return(da);
}

float	Cylinder::Angle2Point(Point& p)
{
	Vect tmp=p-cen;
	float tmp2=tmp.inner(axis)/(tmp.length()*axis.length());		
	if(tmp2>1)tmp2=1;
	else if(tmp2<-1)tmp2=-1;
	float da=acos(tmp2);
	return(da);
}


float	Cylinder::Angle2Center(Cylinder *next)
{
	Point p=next->Center();
	return(Angle2Point(p));
}


float	Cylinder::AxisAngleDiff(Cylinder *next)
{
	Vect v=next->Axis();
	return(Angle2Vect(v));
}

float	Cylinder::PlaneAngleDiff(Cylinder * next)
{
	Vect axis2=next->Axis();	
	Vect cencon=next->Center()-cen;	cencon.normalize();
	Vect n1=axis.cross(cencon);		n1.normalize();
	Vect n2=cencon.cross(axis2);	n2.normalize();
	float tmp=n1.inner(n2);
	if(tmp>1)tmp=1;
	else if(tmp<-1)tmp=-1;
	float da=acos(tmp);
	// now determine the sign
	float sign=n1.cross(n2).inner(cencon);
	if(sign<0) da*=-1;
	return(da);
}

bool Cylinder::if_merge(Cylinder*next, float da_limit, float distance_limit)
{
	if (this == next )
		return false;
	
	float	da1 = Angle2Center(next);
	//printf("\tda2= %f",da2*180./PI);

	if (da1 > da_limit && da1< PI-da_limit ) // ignore directionality
		return false;

	float	da2 = next->Angle2Center(this);
	//printf("\nda1= %f",da1*180./PI);

	if (da2 > da_limit && da2< PI-da_limit ) // ignore directionality) 
		return false;

	float	td = CenterDistance(next);
	//printf("\ttd= %f",td);

	if (fabs((td - (Length() * cos(da1) + next->Length() * cos(da2)) / 2.)) > distance_limit) 
		return false;
	//printf("\tcould be merged\n",td);
	return true;
}


bool Cylinder::merge(Cylinder * next)
{
	float len2=next->Length();
	Point cen2=next->Center();
	Vect axis2=next->Axis();
	Vect cencon=cen2-cen;
	axis=(axis*length+axis2*len2)/(length+len2);
	cen=cen+cencon*len2/(length+len2);	// may need better algorithm
	length=Angle2Vect(axis)*length/2+next->Angle2Vect(axis)*len2/2+cencon.angle(axis)*cencon.length();
	
	return true;
}


/***************************************************************************************
		Cylinders class
****************************************************************************************/
int Cylinders::size()
{
	return cyls.size();
}

Cylinders Cylinders::copy()
{
	Cylinders ret;
	for(int i=0;i<size();i++){
		Cylinder* tc=new Cylinder(cyls[i]->Center(),cyls[i]->Axis(),cyls[i]->Length(),cyls[i]->Radius(),cyls[i]->Weight());
		tc->setStartID(cyls[i]->StartID());	tc->setEndID(cyls[i]->EndID());
		tc->setLabel(cyls[i]->Label());
		ret.add(tc);
	}
	return ret;
}

Cylinders Cylinders::subset(int* list, int n)
{
	Cylinders ret;
	if(n>size()){
		fprintf(stderr,"Cylinders::subset(int* list, int n): invalid list size=%d, should <=%d\n",n,size());
		exit(1);
	}
		
	for(int i=0;i<n;i++){
		if(list[i]>=size()){
			fprintf(stderr,"Cylinders::subset(int* list, int n): invalid index defined in list[%d]=%d\n",i,list[i]);
			exit(1);
		}
		else if (list[i]>=0) { // ignore the negative entries
			Cylinder* tc=new Cylinder(cyls[list[i]]->Center(),cyls[list[i]]->Axis(),cyls[list[i]]->Length(),
				cyls[list[i]]->Radius(),cyls[list[i]]->Weight());
			tc->setStartID(cyls[list[i]]->StartID());	tc->setEndID(cyls[list[i]]->EndID());
			tc->setLabel(cyls[list[i]]->Label());
			ret.add(tc);
		}
	}
	return ret;
}

bool Cylinders::add(Cylinder * cyl){
	cyls.push_back(cyl);	
	return true;
}

bool Cylinders::ReadListfile(FILE * in) {
	float	xc, yc, zc, q0, q1, q2, q3, rad, len;
	int	count = 0;
	while (!feof(in)) {
		if (fscanf(in, "%f %f %f %f %f %f %f %f %f\n", &xc, &yc, &zc, &q1, &q2, &q3, &q0, &rad, &len) != 9) {
			cerr << "ERROR:\tinput file is in wrong format\n";
			return false;	//exit(1);
		}

		Quater	tmp(q1, q2, q3, q0);
		Euler tmp2=(Euler(-PI/2.,0,0)*tmp.toEuler()).inverse();
		Vect tmp3=Vect(0,0,1).rotate(tmp2);
		Cylinder * cyl = new Cylinder(Point(xc,yc,zc), tmp3, len, rad);
		cyls.push_back(cyl);
	}
	return true;
}

bool Cylinders::ReadListfile(const char* in) {
	FILE * fp=stdin;
	if(!strcmp(in,"-")) {
		fp=stdin;
	} 
	else {
		if(!(fp=fopen(in,"rb"))) {
			cerr << "\"Cylinders::ReadListfile\" cannot open file \""<<in<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->ReadListfile(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::WriteListfile(FILE * out)
{
	for (int i = 0; i < cyls.size(); i++) {
		cyls[i]->print(i, out);
	}
	return true;
}

bool Cylinders::WriteListfile(const char* out)
{
	FILE * fp=stdout;
	if(!strcmp(out,"-")) {
		fp=stdout;
	} 
	else {
		if(!(fp=fopen(out,"wb"))) {
			cerr << "\"Cylinders::WriteListfile\" cannot open file \""<<out<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->WriteListfile(fp);
	fclose(fp);
	return(ret);
}


bool Cylinders::ReadInventor(const char* in)
{
	FILE * fp=stdin;
	if(!strcmp(in,"-")) {
		fp=stdin;
	} 
	else {
		if(!(fp=fopen(in,"rb"))) {
			cerr << "\"Cylinders::ReadInventor\" cannot open file \""<<in<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->ReadInventor(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::ReadInventor(FILE * in)	{
#ifdef OPENIV
	SoInput SceneInput;
	SceneInput.setFilePointer(in);
		
	static bool init=true;
	if(init){
		SoDB::init();
		init=false;
	}

	SoSeparator *rootNode = SoDB::readAll(&SceneInput);
	if(rootNode == NULL)
	{
		cerr <<"\"Cylinders::ReadInventor\" has problems when reading file\n";
		return false;
	} 

	SceneInput.closeFile();
	int n = rootNode->getNumChildren();
	if(n==0)
	{
		cerr << "\"Cylinders::ReadInventor\": no children in \""<<in<<"\"\n";
		return false;
	} 

	for(int i=0;i<n;i++) {
		SoSeparator     *helix          = (SoSeparator *)rootNode->getChild(i);
		SoLabel         *label          = (SoLabel *)           helix->getChild(0);
		SoBaseColor     *colorN         = (SoBaseColor *)       helix->getChild(1);
		SoTransform     *transform      = (SoTransform *)       helix->getChild(2);
		SoCylinder      *cylinder       = (SoCylinder *)        helix->getChild(3);

 		float	xc, yc, zc, Q, qx, qy, qz, rad, len;
		const SbVec3f VT=transform->translation.getValue();
		VT.getValue(xc,yc,zc);
                
		SbVec3f VR;
		transform->rotation.getValue(VR,Q);
		VR.getValue(qx,qy,qz);
		rad = cylinder->radius.getValue();
        len = cylinder->height.getValue();
		
		Quater	tmp(qx, qy, qz, Q);
		Euler tmp2=(Euler(-PI/2.,0,0)*tmp.toEuler()).inverse();
		//Euler tmp2=Euler(-PI/2.,0,0)*tmp.toEuler();
		Vect tmp3=Vect(0,0,1).rotate(tmp2);
		Cylinder * cyl = new Cylinder(Point(xc,yc,zc), tmp3, len, rad);				
		cyls.push_back(cyl);
	}
	return true;
#else
	cerr << "\"Cylinders::ReadInventor(FILE* in)\" is not supported on non-SGI platforms\n";
	return(false);
#endif
}

bool Cylinders::WriteInventor(FILE * out)
{
	fprintf(out, "#Inventor V2.0 ascii\n\nSeparator {\n");
	int	id = 0;
	for (int i = 0; i < cyls.size(); i++) {
		cyls[i]->toInventor(out, id++);
	}
	fprintf(out, "}\n");
	return true;
}

bool Cylinders::WriteInventor(const char* out)
{
	FILE * fp=stdout;
	if(!strcmp(out,"-")) {
		fp=stdout;
	} 
	else {
		if(!(fp=fopen(out,"wb"))) {
			cerr << "\"Cylinders::WriteInventor\" cannot open file \""<<out<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->WriteInventor(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::WriteVRML(FILE * out)
{
	fprintf(out, "#VRML V1.0 ascii\n\nSeparator {\n");
	int	id = 0;
	for (int i = 0; i < cyls.size(); i++) {
		cyls[i]->toInventor(out, id++);
	}
	fprintf(out, "}\n");
	return true;
}

bool Cylinders::WriteVRML(const char* out)
{
	FILE * fp=stdout;
	if(!strcmp(out,"-")) {
		fp=stdout;
	} 
	else {
		if(!(fp=fopen(out,"wb"))) {
			cerr << "\"Cylinders::WriteVRML\" cannot open file \""<<out<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->WriteVRML(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::ReadDejaVu(const char* in)
{
	FILE * fp=stdin;
	if(!strcmp(in,"-")) {
		fp=stdin;
	} 
	else {
		if(!(fp=fopen(in,"rb"))) {
			cerr << "\"Cylinders::ReadDejaVu\" cannot open file \""<<in<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->ReadDejaVu(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::ReadDejaVu(FILE * in)	
{
	float	xb,yb,zb,xe,ye,ze, xc, yc, zc, q0, q1, q2, q3;
	char s[256], dummy[32];
	while ((fgets(s,256,in)!=NULL)) {
		if(strncmp(s,"ALPHA",5)==0){
			if (sscanf(s+5, "%s %s %s %s %f %f %f %f %f %f\n", dummy,dummy,dummy,dummy, &xb, &yb, &zb, &xe, &ye, &ze) != 10) {
				cerr << "ERROR:\tinput file is in wrong format\n";
				return false;	//exit(1);
			}
			//printf("%f %f %f %f %f %f\n",xb,yb,zb,xe,ye,ze);
			xc=(xb+xe)/2;	yc=(yb+ye)/2;	zc=(zb+ze)/2;

			Vect p(xe-xb,ye-yb,ze-zb);
			Cylinder * cyl = new Cylinder(Point(xc,yc,zc),p,p.length(),2.5);		
			cyls.push_back(cyl);
		}
	}
	return true;		
}

bool Cylinders::WriteDejaVu(FILE *out, const char * mol, const char * note, const char * pdb)
{
	fprintf(out, "!\n!  ===  %s\n!\n",mol);	
	fprintf(out, "MOL   %s\nNOTE  %s\nPDB   %s\n!\n", mol, note, pdb);
	int	id = 0;
	for (int i = 0; i < cyls.size(); i++) {
		cyls[i]->toDejaVu(out, id);
		++id;
	}
	fprintf(out, "ENDMOL\n");
	return true;
}

bool Cylinders::WriteDejaVu(const char * out, const char * mol, const char * note, const char * pdb)
{
	FILE * fp=stdout;
	if(!strcmp(out,"-")) {
		fp=stdout;
	} 
	else {
		if(!(fp=fopen(out,"wb"))) {
			cerr << "\"Cylinders::WriteDejaVu\" cannot open file \""<<out<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->WriteDejaVu(fp,mol,note,pdb);	
	fclose(fp);
	return(ret);
}	

bool Cylinders::ReadCosec(const char* in)
{
	FILE * fp=stdin;
	if(!strcmp(in,"-")) {
		fp=stdin;
	} 
	else {
		if(!(fp=fopen(in,"rb"))) {
			cerr << "\"Cylinders::ReadCosec\" cannot open file \""<<in<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->ReadCosec(fp);
	fclose(fp);
	return(ret);
}

bool Cylinders::ReadCosec(FILE * in)	
{
	float	xb,yb,zb,xe,ye,ze, length;
	int resnum,startid,endid;
	char s[256];
	while ((fgets(s,256,in)!=NULL)) {
		if(strlen(s)>7 && s[6]=='h'){
			if (sscanf(s+8, "%f %f %f %f %f %f %f %d %d - %d\n", &xb, &yb, &zb, &xe, &ye, &ze, &length, &resnum, &startid, &endid) != 10) {
				cerr << "Cylinders::ReadCosec(FILE * in):\tinput file is in wrong format\n";
				return false;	//exit(1);
			}
			//printf("%f %f %f %f %f %f\n",xb,yb,zb,xe,ye,ze);

			Vect p(xe-xb,ye-yb,ze-zb);
			Cylinder * cyl = new Cylinder(Point((xb+xe)/2,(yb+ye)/2,(zb+ze)/2),p,length,2.5);
			cyl->setStartID(startid);	cyl->setEndID(endid);			
			cyls.push_back(cyl);
			printf("%s",s);
		}
	}
	return true;		
}


bool Cylinders::WriteCosec(FILE * out, const char *prog, const char *pdbid, const char *header, const char *compnd, const char *source, const char *author)
{	
	int i,j,n=cyls.size();
	if(n<2)return false;
	
	// header section
	time_t t = time(NULL);
	fprintf(out,"program  %s\n",prog);
	fprintf(out,"date     %s",asctime(localtime(&t)));
	fprintf(out,"pdbid    %s\n",pdbid); 		
	fprintf(out,"header   %s\n",header); 		
	fprintf(out,"compnd   %s\n",compnd);
	fprintf(out,"source   %s\n",source); 		
	fprintf(out,"author   %s\n",author);
	fprintf(out,"chain    \n");
	fprintf(out,"n_sse       %d\nn_helix     %d\nn_strand    0\nn_elm       %d\n",n,n,n*(n-1)/2);
	
	// helices themselves
	fprintf(out,"  # type   X,   Y,   Z-init          X,   Y,   Z-term       length  naa      seq\n");   
	float	xe1, xe2, ye1, ye2, ze1, ze2;
	for( i=0;i<n;i++)	cyls[i]->toCosec(out,i+1);
	
	// inter-helices geometry
	fprintf(out,"//\nsse1 sse2 type     d        theta1     theta2       phi\n");   
	for( i=0;i<n;i++){
		for( j=i+1;j<n;j++){
			float d=cyls[i]->CenterDistance(cyls[j]);
			float theta1=PI-cyls[i]->Angle2Center(cyls[j]);
			float theta2=cyls[j]->Angle2Center(cyls[i]);
			//float phi=cyls[i]->AxisAngleDiff(cyls[j]);
			float phi=cyls[i]->PlaneAngleDiff(cyls[j]);
			fprintf(out,"%4d %4d  hh %10.6f %10.6f %10.6f %10.6f\n",i+1,j+1,d,theta1,theta2,phi);
		}
	}
	return true;		
}

bool Cylinders::WriteCosec(const char * out, const char *prog, const char *pdbid, const char *header, const char *compnd, const char *source, const char *author)
{
	if(cyls.size()<2)return false;
	
	FILE * fp=stdout;
	if(!strcmp(out,"-")) {
		fp=stdout;
	} 
	else {
		if(!(fp=fopen(out,"wb"))) {
			cerr << "\"Cylinders::WriteCosec\" cannot open file \""<<out<<"\"\n";
			return false;
		}
	}
	
	bool ret=this->WriteCosec(fp,prog,pdbid,header,compnd,source,author);	
	fclose(fp);
	return(ret);
}	

float Cylinders::OrientationDiff(Cylinders& comp, bool ignoredir, bool uselengthweight)	// sum of individual cylinders angle diff
{
	if(size()!=comp.size()){
		cerr<<"Cylinders::OrientationDiff(Cylinders& comp): number of cylinders are different: "<<size()<<" != "<<comp.size()<<endl;
		exit(1);
	}
	float ret=0, totallength=0;
	for(int i=0; i<size(); i++) {
		float tl=cyls[i]->Length()+comp[i]->Length();
		float diff=cyls[i]->AxisAngleDiff(comp[i]);
		//if(diff<0 || diff >PI/2.) printf("1 Bad angle diff = %f\ttl=%f\n", diff,tl);
		if(ignoredir && diff>PI/2.)diff=PI-diff;
		//if(diff<0 || diff >PI/2.) printf("2 Bad angle diff = %f\tt1=%f\n", diff,tl);
		if(uselengthweight)diff*=tl;
		totallength+=tl;
		ret+=diff;
	}
	if(uselengthweight)	ret/=totallength;
	else				ret/=size();
	//if(ret<0 || ret >PI/2.) printf("  Bad angle ret  = %f\n", ret);
	return ret;
}

void Cylinders::print()
{
	for(int i=0;i<cyls.size(); i++) {
		cyls[i]->print(i,stdout);
	}
}

void Cylinders::mirror()
{
	for(int i=0;i<cyls.size(); i++) {
		cyls[i]->mirror();
	}
}
		
void Cylinders::Transform(Quater& rot, Vect& trans, Vect& center)	// transform: rotate then translate
{
	for(int i=0;i<cyls.size(); i++) {
		cyls[i]->Transform(rot,trans,center);
	}
}
		
void Cylinders::Transform(Euler& rot, Vect& trans, Vect& center)	// transform: rotate then translate
{
	for(int i=0;i<cyls.size(); i++) {
		cyls[i]->Transform(rot,trans,center);
	}
}
		
bool Cylinders::SortLength(){
	sort(cyls.begin(), cyls.end(), CylLenComp());
	return true;
}

bool Cylinders::SortWeight(){
	sort(cyls.begin(), cyls.end(), CylWeightComp());
	return true;
}

/*#############################################################################################
  ######																				#######
  ######	fit the object with an cylinder using Jacobi rotation matrix method to		#######
  ######	diagonize the symmetric second moment 3x3 matrix							#######
  ######																				#######		
  #############################################################################################*/

void	ObjectStat::FitEllipsoid(bool sort)
{
	/* determine eigenvalues and eigenvectors */
	jacobi(moments, eigenvalue, eigenvector);

	/* normalize eigenvalues */
	this->Norm();

	if (sort) {
		/* order eigenvalues and eigenvectors and check multiplicity */
		eigsrt(eigenvalue, eigenvector);

		/* calculate aspect ratio from the 3 ellipsoid axes */
		AspectRatio(eigenvalue, aspectratio);

		/* calculate the two angles for the longest axis */
		this->ComputeLongAxis();

		len		= 2 * eigenvalue[0];
		rad		= eigenvalue[1];
	}
}


#define SMALL (1e-7)
void	jacobi(float nu[][3], float d[], float v[][3])
{
	int	j, iq, ip, i, n, nrot;
	float	tresh, theta, tau, t, sm, s, h, g, c;
	float   a[3][3], b[3], z[3];

	n = 3;
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			a[i][j] = nu[i][j];
		}
	}

	for (ip = 0; ip < n; ip++) {
		for (iq = 0; iq < n; iq++) {
			v[ip][iq] = 0.0;
		}
		v[ip][ip] = 1.0;
	}

	for (ip = 0; ip < n; ip++) {
		b[ip] = a[ip][ip];
		d[ip] = b[ip];
		z[ip] = 0.0;
	}
	nrot = 0;

	for (i = 0; i < 50; i++) {
		sm = 0.0;
		for (ip = 0; ip < n - 1; ip++) {
			for (iq = ip + 1; iq < n; iq++) {
				sm = sm + fabs(a[ip][iq]);
			}
		}

		if (sm < SMALL ) 
			return;

		if (i < 3) 
			tresh = 0.2 * sm / n / n;
		else	
			tresh = 0.0;
		for (ip = 0; ip < n - 1; ip++) {
			for (iq = ip + 1; iq < n; iq++) {
				g = 100.0 * fabs(a[ip][iq]);
				if (i > 3 && (fabs((fabs(d[ip]) + g) - fabs(d[ip])) < 0.5e-15) && (fabs((fabs(d[iq]) +
				    g) - fabs(d[iq])) < 0.5e-15))
					a[ip][iq] = 0.0;
				else if (fabs(a[ip][iq]) > tresh) {
					h = d[iq] - d[ip];
					if (fabs((fabs(h) + g) - fabs(h)) <0.5e-15) 
						t = a[ip][iq] / h;
					else {
						theta = 0.5 * h / a[ip][iq];
						t = 1.0 / (fabs(theta) + sqrt(1.0 + theta * theta));
						if (theta < 0.) 
							t = -t;
					}
					c = 1.0 / sqrt(1 + t * t);
					s = t * c;
					tau = s / (1.0 + c);
					h = t * a[ip][iq];
					z[ip] -= h;
					z[iq] += h;
					d[ip] -= h;
					d[iq] += h;
					a[ip][iq] = 0.0;
					for (j = 0; j < ip; j++) {
						g = a[j][ip];
						h = a[j][iq];
						a[j][ip] = g - s * (h + g * tau);
						a[j][iq] = h + s * (g - h * tau);
					}
					for (j = ip + 1; j < iq; j++) {
						g = a[ip][j];
						h = a[j][iq];
						a[ip][j] = g - s * (h + g * tau);
						a[j][iq] = h + s * (g - h * tau);
					}
					for (j = iq + 1; j < n; j++) {
						g = a[ip][j];
						h = a[iq][j];
						a[ip][j] = g - s * (h + g * tau);
						a[iq][j] = h + s * (g - h * tau);
					}
					for (j = 0; j < n; j++) {
						g = v[j][ip];
						h = v[j][iq];
						v[j][ip] = g - s * (h + g * tau);
						v[j][iq] = h + s * (g - h * tau);
					}
					nrot ++;
				}
			}
		}
		for (ip = 0; ip < n; ip++) {
			b[ip] += z[ip];
			d[ip]  = b[ip];
			z[ip]  = 0.0;
		}
	}
	fprintf(stderr, "Stop in subroutine \"jacobi\":\t");
	fprintf(stderr, "should not take 50 iterations! sm = %g\n", sm);
	exit(1);
}


void	eigsrt(float d[], float v[][3])	 // from large to small
{										// vector i= eigenvector[0-2][i]			
	int	k, i, j, n;
	float   p;

	n = 3;

	for ( i = 0; i < n - 1; i++) {
		k = i;
		p = d[i];
		for ( j = i + 1; j < n; j++) {
			if ( d[j] > p ) {
				k = j;
				p = d[j];
			}
		}
		if ( k != i ) {
			d[k] = d[i];
			d[i] = p;
			for ( j = 0; j < n; j++) {
				p = v[j][i];
				v[j][i] = v[j][k];
				v[j][k] = p;
			}
		}
	}
}

void	ObjectStat::ComputeLongAxis()
{
	/*Vect p(eigenvector[0][0], eigenvector[1][0], eigenvector[2][0]);
	p.normalize();
	alt = acos(p.Z());
	az  = atan2(p.Y(), p.X()) + PI / 2;	// using EMAN convention ZX, shperical coordinates use ZY convention	*/
	axis.setXYZ(eigenvector[0][0], eigenvector[1][0], eigenvector[2][0]);
	axis.normalize();
}

void	ObjectStat::Norm() // normalize the eigenvalues so that the volume PI*eigen[0]*eigen[1]*eigen[2]==volume
{						// the same time change unit of eigenvalues	from m^2 to m	
	int	i;
	for (i = 0; i < 3; i++) {
		if (eigenvalue[i] > 0) 
			eigenvalue[i] = sqrt(eigenvalue[i]);
		else 
			eigenvalue[i] = 0;
	}
	float	cylvol = 2*PI*eigenvalue[1]*eigenvalue[2]*eigenvalue[0];
	if (cylvol > 0) {
		float   factor = pow((float)(volume / cylvol), (float)(1.0 / 3.0));
		for (i = 0; i < 3; i++) { 
			eigenvalue[i] *= factor;
		}
	}
}


void	AspectRatio(float evalue[3], float ratio[3])
{
	int	i, j;
	float	ev[3];
	float   temp;

	for (i = 0; i < 3; i++) 
		ratio[i] = 0;
	if (evalue[1] > 0)
		ratio[0] = evalue[0] / evalue[1];
	if (evalue[2] > 0) {
		ratio[1] = evalue[0] / evalue[2];
		ratio[2] = evalue[1] / evalue[2];
	}

	/*for (i=0; i<3; i++) 
       ev[i]=sqrt(evalue[i]);
        
   for (i=0;i<3;i++) {
        for (j=i;j<3;j++) {
	   		if (ev[i]<ev[j]) {
	       		temp=ev[j];
	       		ev[j]=ev[i];
	       		ev[i]=temp;
	   		}
		}
   }
       
   ratio[0]=(ev[0])/(ev[1]);
   ratio[1]=(ev[0])/(ev[2]);
   ratio[2]=(ev[1])/(ev[2]);
   */
}


void	ObjectStat::SetObject(EMData *map, const vector < Point * > &obj)
{
	this->volume = 0;
	this->mass = 0;
	this->xc = 0;
	this->yc = 0;
	this->zc = 0;
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			moments[i][j] = 0;
		}
	}

	float	apix = map->Pixel();
	float	xorigin=map->getXorigin(), yorigin=map->getYorigin(), zorigin=map->getZorigin(); 
	int	nx = map->xSize(), 	ny = map->ySize(), 	nz = map->zSize();
	float	*dp = map->getDataRO();
	for (int j = 0; j < obj.size(); j++) {
		float	x = apix*obj[j]->X() + xorigin;	
		float	y = apix*obj[j]->Y() + yorigin;
		float	z = apix*obj[j]->Z() + zorigin;

		float	dpvalue = dp[round(obj[j]->Z())*nx*ny+round(obj[j]->Y())*nx+round(obj[j]->X())];
		volume++;
		mass += dpvalue;
		xc += dpvalue * x;
		yc += dpvalue * y;
		zc += dpvalue * z;
		moments[0][0] += dpvalue * x * x;
		moments[0][1] += dpvalue * x * y;
		moments[0][2] += dpvalue * x * z;
		moments[1][1] += dpvalue * y * y;
		moments[1][2] += dpvalue * y * z;
		moments[2][2] += dpvalue * z * z;
	}
	map->doneData();
	volume *= apix * apix * apix;

	xc /= mass;	
	yc /= mass;	
	zc /= mass;
	moments[0][0] = moments[0][0] / mass - xc * xc;
	moments[0][1] = moments[0][1] / mass - xc * yc;
	moments[0][2] = moments[0][2] / mass - xc * zc;
	moments[1][1] = moments[1][1] / mass - yc * yc;
	moments[1][2] = moments[1][2] / mass - yc * zc;
	moments[2][2] = moments[2][2] / mass - zc * zc;
	moments[1][0] = moments[0][1];
	moments[2][0] = moments[0][2];
	moments[2][1] = moments[1][2];

	this->FitEllipsoid();
}


void	ObjectStat::SetObject(EMData *map)
{
	this->volume = 0;
	this->mass = 0;
	this->xc = 0;
	this->yc = 0;
	this->zc = 0;
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			moments[i][j] = 0;
		}
	}

	float	apix = map->Pixel();
	float	origx=map->getXorigin(), origy=map->getYorigin(), origz=map->getZorigin(); 
	int	nx = map->xSize(), 	ny = map->ySize(), 	nz = map->zSize();
	float	*dp = map->getDataRO();
	for (int k = 0; k < nz; k++) {
		float	z = apix*k + origz;
		for (int j = 0; j < ny; j++) {
			float	y = apix*j + origy;
			for (int i = 0; i < nx; i++) {
				float	x = apix*i + origx;
				float	dpvalue = dp[k*nx*ny+j*nx+i];
				volume++;
				mass += dpvalue;
				xc += dpvalue * x;
				yc += dpvalue * y;
				zc += dpvalue * z;
				moments[0][0] += dpvalue * x * x;
				moments[0][1] += dpvalue * x * y;
				moments[0][2] += dpvalue * x * z;
				moments[1][1] += dpvalue * y * y;
				moments[1][2] += dpvalue * y * z;
				moments[2][2] += dpvalue * z * z;
			}
		}
	}
	map->doneData();
	volume *= apix * apix * apix;

	xc /= mass;	
	yc /= mass;	
	zc /= mass;
	moments[0][0] = moments[0][0] / mass - xc * xc;
	moments[0][1] = moments[0][1] / mass - xc * yc;
	moments[0][2] = moments[0][2] / mass - xc * zc;
	moments[1][1] = moments[1][1] / mass - yc * yc;
	moments[1][2] = moments[1][2] / mass - yc * zc;
	moments[2][2] = moments[2][2] / mass - zc * zc;
	moments[1][0] = moments[0][1];
	moments[2][0] = moments[0][2];
	moments[2][1] = moments[1][2];

	bool sort = false;
	this->FitEllipsoid(sort);	// not to sort so that the order is right
}


void	ObjectStat::EigenSortTo(ObjectStat*obj)
{
	int	i, j, n;
	float	p;

	n = 3;

	for ( i = 0; i < n; i++) {
		float	min = FLT_MAX;
		int	minid = i;
		p = obj->eigenvalue[i];
		for ( j = i; j < n; j++) {
			if ( fabs(eigenvalue[j] - p) < min ) {
				minid  	 = j;
				min 	 = fabs(eigenvalue[j] - p);
			}
		}
		if ( minid != i ) {
			p = eigenvalue[minid];
			eigenvalue[minid] = eigenvalue[i];
			eigenvalue[i] = p;

			for ( j = 0; j < n; j++) {
				p = eigenvector[j][i];
				eigenvector[j][i] = eigenvector[j][minid];
				eigenvector[j][minid] = p;
			}
		}
	}


}

void	ObjectStat::print()
{
	printf("Object centered at (%.2f,%.2f,%.2f) with mass= %g, volume= %g, length= %g\n", xc, yc, zc, mass, volume, len);
	printf("Moments:\t\t\tEigenValue:\tEigenVector:\n");
	printf("%8.2f %8.2f %8.2f\t%7.2f\t\t%7.2f %7.2f %7.2f\n", moments[0][0], moments[0][1], moments[0][2],
	    eigenvalue[0], eigenvector[0][0], eigenvector[1][0], eigenvector[2][0]);
	printf("%8.2f %8.2f %8.2f\t%7.2f\t\t%7.2f %7.2f %7.2f\n", moments[1][0], moments[1][1], moments[1][2],
	    eigenvalue[1], eigenvector[0][1], eigenvector[1][1], eigenvector[2][1]);
	printf("%8.2f %8.2f %8.2f\t%7.2f\t\t%7.2f %7.2f %7.2f\n", moments[2][0], moments[2][1], moments[2][2],
	    eigenvalue[2], eigenvector[0][2], eigenvector[1][2], eigenvector[2][2]);
	printf("Aspect ratio:\t%8.2g %8.2g %8.2g\n\n", aspectratio[0], aspectratio[1], aspectratio[2]);
}

inline float radprofile(float r, int type)
{ // r in angstrom
	float	ret = 0;
	if (type == 0) {	// pure Gaussian falloff
		r /= 2;
		ret = exp(-r * r);
	} else if (type == 1) {	// pure Gaussian falloff + negative dip, so mean is 0
		r /= 2;
		ret = (1 - r * r / 4) * exp(-r * r / 4);
	} else if (type == 2) {
		// polynomial fitting to the radial profile of real helix density
		// f=a0*x^n+a1+x^(n-1)+ ... +a[n-1]*x+an
		//float an[11]={2.847024584977009e-10,-3.063997224364090e-08,1.418801040660860e-06,-3.678676414383996e-05,5.804871622801710e-04,-5.640340018430164e-03,3.208802421493864e-02,-9.068475313823952e-02,7.097329559749284e-02,-9.993347339658298e-02,1.000000000000000e+00}; 

		// now the fitting to the original profile		
		static float	an[15] = {
			-3.9185246832229140e-16, 3.3957205298900993e-14, 2.0343351971222658e-12, -4.4935965816879751e-10,
			 3.0668169835080933e-08, -1.1904544689091790e-06, 2.9753088549414953e-05, -4.9802112876220150e-04,
			 5.5900917825309360e-03, -4.0823714462925299e-02, 1.8021733669148599e-01, -4.0992557296268717e-01,
			 3.3980328566901458e-01, -3.6062024812411908e-01, 1.0000000000000000e+00
		};

		ret = an[0];
		for (int i = 1; i < 15; i++) {
			ret = ret * r + an[i];
		}
	}
	return ret;
}

EMData *modelHelix(EMData *in, float alt, float az, int type, float len, int x0, int y0, int z0, bool reset)
// symthesize model alpha helix, len is Angstrom, default to 2 turns, alt, az in radian
{
	EMData * cyl = in;
	int	nx = cyl->xSize();
	int	ny = cyl->ySize();
	int	nz = cyl->zSize();
	float	apixel = cyl->Pixel();
	if (reset) 
		cyl->zero();
	float * d = cyl->getData();
	int	ps = (int)(len / apixel);		//len in pixel
	int	ps2 = (int)(len / apixel / 2.0);	//half len in pixel

	if (x0 < 0 || x0 >= nx)
		x0 = nx / 2;
	if (y0 < 0 || y0 >= ny)
		y0 = ny / 2;
	if (z0 < 0 || z0 >= nz)
		z0 = nz / 2;

	for (int k = -ps; k <= ps; k++) {
		for (int j = -ps; j <= ps; j++) {
			for (int i = -ps; i <= ps; i++) {
				int	l = (i + x0) + (j + y0)*nx + (k + z0)*nx*ny;
				if (l >= 0 && l < nx * ny * nz) {
					float	x = i*cos(az)         + j*sin(az);
					float	y = -i*sin(az)*cos(alt) + j*cos(alt)*cos(az) + k*sin(alt);
					float	z = i*sin(alt)*sin(az)-j*sin(alt)*cos(az) + k*cos(alt);
					float	dl = hypot(x*apixel, y*apixel);
					if ((z < ps2) && (z > -ps2) && dl < ps) 
						d[l] += radprofile(dl, type);
				}
			}
		}
	}
	cyl->doneData();
	cyl->update();
	return cyl;
}


void	threshHold(EMData*map, float thresh)
{
	int	nx = map->xSize(), ny = map->ySize(), nz = map->zSize();
	float	norm = map->Max()-thresh;
	float	*p = map->getData();
	for (int i = 0; i < nx * ny * nz; i++) {
		if (p[i] > thresh)	
			p[i] = (p[i] - thresh) / norm;
		else	
			p[i] = 0;
	}
	map->doneData();
	map->update();
}


