// ptcltrace.C	Copyright 2001 Baylor College of Medicine
// This will examine how particle classification changes with iteration

//N ptcltrace
//F Trace how classification of a single particle changes
//T LT
//1
//P [ptcl=<n>[,<n>]]	Particle number to evaluate (range 0-n)
//P [iter=[<n0>,<n1>][<n0>-<n1>]]	Iterations to evaluate (range 1-n)
//P [sym=<cn>]	Specify symmetry of model only for Cn symmetries, NOT Dn 
//P [simple]	Short form report
//P [zap=<maxang>]	Mark particles with a mean error that's too large
//U ptcltrace ptcl=5
//U ptcltrace iter=5,8 zap=5
//D This program is used to analyze particle.log and determine how
//E consistent the orientation assignment for each particle is. In addition
//E it can create a start.exclude file to exclude 'bad' particles from 
//E further rounds of refinement.
//D The per-particle analysis compares the assigned Euler angles of each
//E particle through several refinement iterations. The zap= option is
//E used to specify the maximum allowable average Euler angle variation
//E for each particle. Any particle which is not classified consistently
//E within this angle will be added to the start.exclude file and will not
//E be used in further rounds of refinement.
//D Iteration 0 is special. It represents the Euler angles stored in the particle
//E headers. This can be used within EMAN for testing with simulated data, or
//E with outside software that puts angles in the headers.

#include <stdio.h>
#include <string.h>
#include <time.h>
#include "util.h"
#include "EMData.h"
#include "portable_fileio.h"

#define MAXIT	100

inline float dremfl(float x,float y) { return x-y*floor(x/y); }

int main(int argc,char *argv[]) {
int i,j,j1,j2,j3,k,n,nit=0,dsym=1;
int first=-1,last=-1,firsti=0,lasti=-1,verbose=1;
float symang = PI*2.0, zap = -1.0;
float v;
char useit[MAXIT];

appinit(argc,argv);

for (i=0; i<MAXIT;i++) useit[i]=1;
useit[0]=0;

for (i=1; i<argc; i++) {
	if (strncmp("ptcl=",argv[i],5)==0) sscanf(&argv[i][5],"%d,%d",&first,&last);
	else if (strncmp("iter=",argv[i],5)==0) {
		for(j=0;j<MAXIT;j++) useit[j]=0;
		j=5;
		j3=-1;
		j1=0;
		while (sscanf(&argv[i][j],"%d%n",&j1,&j2)) {
			j+=j2;
			if (j1>=MAXIT) j1=MAXIT-1;
			if (j3>=0) while (j3<=j1) { useit[j3]=1; j3++; }
			j3=-1;
			if (argv[i][j]=='-') j3=j1;
			else if (argv[i][j]==','||argv[i][j]==0) useit[j1]=1;
			else error(ERR_EXIT,"Invalid range specification");
			if (argv[i][j]==0) break;
			j++;
		}
		if (j3>=0) while (j3<MAXIT) { useit[j3]=1; j3++; }
	}
	else if (strncmp("simple",argv[i],6)==0) verbose=0;	
	else if (strncmp("sym=",argv[i],4)==0) {
		if (argv[i][4]!='c' && argv[i][4]!='C') {
			if (argv[i][4]=='d' || argv[i][4]=='D') dsym=atoi(&argv[i][5]);
			error(ERR_WARN,"D symmetries have no effect, only C symmetries are allowed");
		}
		symang=2.0*PI/atof(&argv[i][5]);
	}
	else if (strncmp("zap=",argv[i],4)==0) {
		zap=atof(&argv[i][4]);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

for (i=0; i<20; i++) printf("%0d ",useit[i]);
printf("\n");

int pos[MAXIT];		// currently hardcoded to 100 iterations max
int nproj[MAXIT];

for (i=0; i<MAXIT; i++) { nproj[i]=0; pos[i]=-1; }
pos[0]=1;

n=fileCount("start.hed");
if (first==-1) { first=0; last=n-1; }
else if (last<first) last=first;

FILE *in = fopen("particle.log","rb");
if (!in) error(ERR_EXIT,"Cannot read particle.log");
char s[256],t[256];

// We only want to use the last iteration with each number, ie if iteration 1 was
// run 3 times, we only use the 3rd
int cit=0;
while (fgets(s,255,in)) {
	if (s[0]=='S') {
		sscanf(s+1," %d %d",&j,&k);
		if (j>MAXIT-1) error(ERR_EXIT,"Sorry, MAXIT exceeded");
		printf("Iteration %d at %s",j,ctime((time_t *)&k));
		pos[j]=portable_ftell(in);
		nit=j>nit?j:nit;
		cit=j;
	}
	if (s[0]=='P') {
		if (cit==0) error(ERR_EXIT,"Old particle.log file without S records.");
		sscanf(s+1," %d",&j);
		if (nproj[cit]<j) nproj[cit]=j;
	}
}
nit++;

// This will copy projections from a previous iteration if the current iteration doesn't have any
j=0;
for (i=1; i<MAXIT; i++) {
	if (nproj[i]) j=nproj[i];
	else nproj[i]=j;
}

printf("Found %d iterations\n",nit-1);
printf("%d projections in the final iteration\n",nproj[nit-1]);

// Read all the detailed data about each particle and each projection
// from the last class with the given iteration number present in the file
int **cat;
float **val;
float **alt;
float **az;

cat=(int **)malloc(nit*sizeof(int *));
val=(float **)malloc(nit*sizeof(float *));
alt=(float **)malloc(nit*sizeof(float *));
az=(float **)malloc(nit*sizeof(float *));
for (i=1; i<nit; i++) {
	if (pos[i]<0) continue;
	cat[i]=(int *)malloc((last-first+2)*sizeof(int));
	val[i]=(float *)malloc((last-first+2)*sizeof(float));
	alt[i]=(float *)malloc((nproj[i]+2)*sizeof(float));
	az[i]=(float *)malloc((nproj[i]+2)*sizeof(float));
	for (j=0; j<=nproj[i]; j++) alt[i][j]=-1.0;
	for (j=0; j<last-first+1; j++) { cat[i][j]=0; val[i][j]=-5.0; }
	portable_fseek(in,pos[i],SEEK_SET);
	while (fgets(s,255,in)) {
		if (s[0]=='S') break;
		if (s[0]=='P') {
			sscanf(s+1," %d%n",&j,&k);
			j--;
			sscanf(s+k+1," %f %f",&alt[i][j],&az[i][j]);
		}
		if (s[0]=='C') {
			sscanf(s+1," %d %d %s %d %d",&j,&j1,t,&j2,&j3);
			sscanf(t," %f",&v);
if (j3!=i) printf("WARNING: pt %d,  %d should be %d\n",j,j3,i);
			if (j>=first && j<=last && j1==0) {
				cat[i][j-first]=j2;
				val[i][j-first]=v;
			}
		}
	}
}

// Try to clean up missing projections
for (i=1; i<nit; i++) {
	if (pos[i]<0) continue;
	for (j=0; j<nproj[i]; j++) {
		if (alt[i][j]<0) {
			for (k=1; k<nit; k++) {
				if (pos[k]<0) continue;
				if (nproj[i]!=nproj[k]) continue;
				if (alt[k][j]>=0) { alt[i][j]=alt[k][j]; az[i][j]=az[k][j]; break; }
			}
		}
	}
}

/*
for (i=0; i<nit; i++) {
	printf("\nIteration %d\n",i);
	for (j=0; j<nproj[i]; j++) {
		printf("%d.\t%1.1f\t%1.1f\n",j,alt[i][j],az[i][j]);
	}
}
*/

// final analysis
FILE *ex = NULL;
if (verbose) {
	if (zap>0) ex=fopen("start.exclude","wb");
	FILE *out=fopen("error.txt","wb");
	for (i=first; i<=last; i++) {
		printf("%d. ********\n",i);
		float sum=0,norm=0;
		for (j=0; j<nit; j++) {
			EMData tmp;
			if (useit[j]==0 || pos[j]<0) continue;
			if (j==0) {
				tmp.readImage("start.hed",i,1);
				if (dsym>1) {
					float sa=dremfl(tmp.az(),(2.*PI)/(float)dsym);
					if (sa>PI/(float)dsym) sa=(2.*PI)/(float)dsym-sa;
					tmp.setRAlign(tmp.alt(),sa);
				}
				printf("%2d -> %d (%1.1f,%1.1f)\t",0,-1,tmp.alt()*180./PI,tmp.az()*180./PI);
			}
			else printf("%2d -> %d (%1.1f,%1.1f)\t",j,cat[j][i-first],alt[j][cat[j][i-first]],az[j][cat[j][i-first]]);
			for (k=j+1; k<nit; k++) {
				if (useit[k]==0 || pos[k]<0) continue;
				if ((j>0&&val[j][i-first]<0) || val[k][i-first]<0) { printf("XXX\t"); continue; }
				float d;
				Euler e1,e2,e3,e4;
				if (j==0) {
					e1.setAngle(tmp.alt(),tmp.az(),0);
				}
				else e1.setAngle(alt[j][cat[j][i-first]]*PI/180.0,az[j][cat[j][i-first]]*PI/180.0,0);
				e2.setAngle(alt[k][cat[k][i-first]]*PI/180.0,az[k][cat[k][i-first]]*PI/180.0,0);
				e3.setAngle(alt[k][cat[k][i-first]]*PI/180.0,az[k][cat[k][i-first]]*PI/180.0-symang,0);
				e4.setAngle(alt[k][cat[k][i-first]]*PI/180.0,az[k][cat[k][i-first]]*PI/180.0+symang,0);
				d=e1.diff(&e2);
				if (e1.diff(&e3)<d) d=e1.diff(&e3);		// these lines check for symmetry-related angles
				if (e1.diff(&e4)<d) d=e1.diff(&e4);
				printf("%1.2f\t",d*180.0/PI);
				sum+=d*180.0/PI;
				norm+=1.0;
			}
			printf("\n");
		}

		if (ex && (norm==0 || sum/norm>zap)) { fprintf(ex,"%d\n",i); printf("###\t"); }
		else if (ex) printf("\t");

		printf("MEAN DIFF = %1.2f\n",norm==0?-1.0:(sum/norm));
		if (out) fprintf(out,"%1.2f\n",norm==0?-1.0:(sum/norm));

	}
	if (out) fclose(out);
	if (ex) fclose(ex);
}
else {
	for (i=first; i<=last; i++) {
		printf("%d.\t",i);
		for (j=0; j<nit; j++) {
			if (useit[j]==0) continue;
			if (cat[i-first]>=0) printf("%d (%1.1f)\t",cat[j][i-first],val[j][i-first]);
			else printf("        \t");
		}
		printf("\n");
	}
}
}
